> {-# OPTIONS -fglasgow-exts -XNoMonomorphismRestriction #-}

> module Mixins where
>
> import Control.Monad.Zipper hiding (leftL, rightL, from, to, (:><:), getv, i, o, vcomp)
> import Control.Monad.State hiding (fix, evalStateT)
> import Control.Monad.Reader hiding (fix)
> import Data.Map
> import Data.List hiding (insert)
> import Control.Monad.Identity hiding (fix)
> import Control.Monad.Error hiding (fix)
> import Zipper
> import Views
> import Mask

> infixl 8 <@>

> runId = runIdentity

> type Id = Identity

> runL = leftL

> type Mixin s = s -> s
>
> fix :: Mixin s -> s
> fix a = a (fix a)

> (<@>) :: Mixin s -> Mixin s -> Mixin s
> a1 <@> a2  = \proceed -> a1 (a2 proceed)

> fib1 :: Mixin (Int -> Int) 
> fib1 proceed n = case n of
>  0 -> 0
>  1 -> 1

> fib2 :: Mixin (Int -> Int)
> fib2 proceed n = case n of
>  10  -> 55
>  _   -> proceed n

> slowfib1, slowfib2 :: Int -> Int
> slowfib1  = fix fib1
> slowfib2  = fix (fib2 <@> fib1)

> memo ::  MonadState (Map Int Int) m => Mixin (Int -> m Int) 
> memo proceed x = 
>   do  m <- get
>       if  member x m 
>           then return (m ! x)
>           else do  y <- proceed x
>                    m' <- get
>                    put (insert x y m')
>                    return y
>
> fib2' :: Monad m => Mixin (Int -> m Int)
> fib2' proceed n = 
>  case n of
>    0 -> return 0
>    1 -> return 1
>    _ -> do  y <- proceed (n-1)
>             x <- proceed (n-2) 
>             return (x+y)

> slowfib2' :: Int -> Int
> slowfib2' = runId . fix fib2'
>
> fastfib :: Int -> Int
> fastfib = evalM empty . fix (memo <@> fib2')
>
> evalM :: s -> StateT s Id a -> a
> evalM s m = runId $ runStateT m s >>= return . fst

> prof :: MonadState Int m => Mixin (a -> m b) 
> prof proceed x = do  c <- get 
>                      put (c + 1)
>                      proceed x

> proffib = evalM 0 . fix (prof <@> fib2')

> infixr 5 <@@>

> (<@@>)  ::  Mixin (a -> t1 (t2 m) b) 
>         ->  Mixin (a -> (t1 :> t2) m b) 
>         ->  Mixin (a -> t1 (t2 m) b)
> c1 <@@> c2 = \proceed x -> c1 (leftL . c2 (rightL . proceed)) x

> fmask  ::  (Monad m, Monad n) 
>        =>  (n :><: m) -> Mixin (a -> n b) -> Mixin (a -> m b)
> fmask v mix proceed = from v . mix (to v . proceed)
>
> mix `usesm` names  = fmask (structure names) mix 

> assertInput :: (Show t, MonadError String m) => (t -> Bool) -> Mixin (t -> m t)
> assertInput f proceed x 
>  | f x       = proceed x
>  | otherwise = throwError ("Assertion failed for " ++ show x ++ "!")
>
> test0 :: Int -> Either String Int
> test0 = fix (assertInput (>= 0) <@> fib2')

> assertOutput ::  (MonadError String m, Show a, Show b) => (b -> Bool) -> Mixin (a -> m b)
> assertOutput f proceed x = 
>  do  r <- proceed x
>      if f r  then return r 
>              else throwError (msg1 ++ show x ++ msg2 ++ show r ++ "!") 
>  where  msg1 = "Assertion failed for input "
>         msg2 = " with output "

> assertDump  ::  (MonadError String m, MonadReader s m, Show a, Show b, Show s) 
>             =>  (b -> Bool) -> Mixin (a -> m b)

> assertDump f proceed x = 
>  catchError  (assertOutput f proceed x) 
>              (\s -> do {m <- ask; throwError (s ++ "\n\n Dump: " ++ show m)})

> data Memo = Memo
> data Prof = Prof
> data Err  = Err

> test :: Int -> ((Either String Int, Int), Map Int Int)
> test = myEval . fix  (    prof              `usesm` Prof
>                      <@>  memo              `usesm` Memo
>                      <@>  fib2')
>
> myEval m = runId $ runTStateT Memo empty $ runTStateT Prof 0 $ runTErrorT Err $ m
